# HTML to ASCII converter by Joe V. Moss (joe@morton.rain.com)
# Copyright (c) 1995 by Joseph V. Moss
#

# These are Defined in HTML 2.0

array set HTtext_map {
	a	{}
	b	*
	cite	{\"}
	code	{\"}
	em	_
	i	_
	kbd	{\"}
	samp	{\"}
	strong	*
	tt	{}
	var	{` '}
}

# These are in common(?) use, but not defined in html2.0

array set HTtext_map {
	strike	{}
	u	_
}

proc HTinit_state {param text} {
	upvar #0 HTarr var
	array set tmp [array get var {[A-Z]*}]
	catch {unset var}
	array set var {
		para		""
		out		""
		pind1		0
		pind2		0
		preflag		0
		nonewline	0
		indent		0
		baseindent	0
		listlvl		0
		bqlvl		0
		DTwidth		20
		LineWidth	79
		TextWidth	75
		Symbols		*+-O=o>:+-+-+-+-
	}
	array set var [array get tmp]
	set var(pllen)	$var(TextWidth)
}

# html tags requiring special processing
# Procs of the form HTtag_<tag> or HTtag_</tag> get called just before
# the text for this tag is displayed.
#   param: The un-interpreted parameter list
#   text:  A pass-by-reference name of the plain text until the next html tag
#          Tag commands may change this to affect what text will be inserted
#          next.

proc HTtag_hmstart {param text} {
	HTinit_state $param $text
}

proc HTtag_/hmstart {param text} {
	HTnewpara 78 0 0 ""
}

proc HTtag_title {param text} {
	upvar $text data
	set data ""
}

proc HTtag_/title {param text} {
	upvar #0 HTarr var
	set var(para) ""
}

proc HTtag_!-- {param text} {
	upvar #0 HTarr var
	if { [regexp -nocase \
			{textindent ([-+]?)([0-9]+) ([-+]?)([0-9]+) --} \
			$param dummy rel1 first rel2 rest] } {
		if { "X$rel1" == "X" } {
			set var(pind1) $first
		} else {
			incr var(pind1) $rel1$first
		}
		if { "X$rel2" == "X" } {
			set var(pind2) $rest
		} else {
			incr var(pind2) $rel1$rest
		}
	}
}

proc HTtag_img {param text} {
	upvar #0 HTarr var
	upvar $text data
	set alt ""
	HMextract_param $param alt
	set alt [HMmap_esc $alt]
	if { "X$alt" != "X" } {
		set data "$alt $data"
	}
}

proc HTtag_h1 {param text} {
	upvar #0 HTarr var
	HTnewpara $var(TextWidth) 0 0 ""
}

proc HTtag_/h1 {param text} {
	upvar #0 HTarr var
	set hdr [HMzap_white [string toupper $var(para)]]
	set len [string length $hdr]
	set spaces [replicate " " [expr ($var(LineWidth)-$len)/2]]
	append var(out) "$spaces$hdr\n"
	append var(out) "$spaces[replicate - $len]\n"
	set var(para) ""
	upvar $text data
	if { [string length [string trim [HMzap_white $data]]] } {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_h2 {param text} {
	upvar #0 HTarr var
	HTnewpara $var(TextWidth) 0 0 ""
}

proc HTtag_/h2 {param text} {
	upvar #0 HTarr var
	set var(para) [string toupper $var(para)]
	upvar $text data
	if { [string length [string trim [HMzap_white $data]]] } {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_h3 {param text} {
	upvar #0 HTarr var
	upvar $text data
	HTnewpara $var(TextWidth) 0 0 ""
}

proc HTtag_/h3 {param text} {
	upvar $text data
	if { [string length [string trim [HMzap_white $data]]] } {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_h4 {param text} {
	upvar #0 HTarr var
	upvar $text data
	HTnewpara $var(TextWidth) 2 2 ""
}

proc HTtag_/h4 {param text} {
	upvar #0 HTarr var
	set var(para) [string toupper $var(para)]
	upvar $text data
	if [string length [string trim [HMzap_white $data]]] {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_h5 {param text} {
	upvar #0 HTarr var
	upvar $text data
	HTnewpara $var(TextWidth) 2 2 ""
}

proc HTtag_/h5 {param text} {
	upvar $text data
	if [string length [string trim [HMzap_white $data]]] {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_h6 {param text} {
	upvar #0 HTarr var
	upvar $text data
	HTnewpara $var(TextWidth) 4 4 ""
}

proc HTtag_/h6 {param text} {
	upvar $text data
	if [string length [string trim [HMzap_white $data]]] {
		uplevel [list HTtag_p $param $text]
	}
}

proc HTtag_br {param text} {
	upvar $text data
	upvar #0 HTarr var
	HTnewpara $var(pllen) $var(pind2) $var(pind2) ""
}

proc HTtag_hr {param text} {
	upvar #0 HTarr var
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
	append var(out) "  [replicate - [expr $var(LineWidth)-4]]\n"
}

proc HTtag_pre {param text} {
	upvar #0 HTarr var
	set var(preflag) 1
	set var(baseindent) 0
	set var(indent) 0
	HTnewpara $var(TextWidth) 0 0 ""
}

proc HTtag_/pre {param text} {
	upvar #0 HTarr var
	upvar $text data
	set var(preflag) 0
	set var(nonewline) 1
	set var(baseindent) 3
	set var(indent) [HTcalcindent]
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
	#if {[string length [HMzap_white $data]] >1 } {
	#	append var(out) "\n"
	#}
	append var(out) "\n"
	#puts stderr "/PRE:$data:"
}

proc HTtag_p {param text} {
	upvar #0 HTarr var
	upvar $text data
	set var(baseindent) 3
	set var(indent) [HTcalcindent]
	set indent $var(indent)
	if { $var(listlvl) && $var(pind2) > $indent } {
		set indent $var(pind2)
	}
	HTnewpara $var(TextWidth) $indent $indent ""
	#puts stderr "P:$data:"
}

proc HTtag_blockquote {param text} {
	upvar #0 HTarr var
	incr var(bqlvl)
	set var(indent) [HTcalcindent]
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
	upvar $text data
	#puts stderr "BLOCKQUOTE:$data:"
}

proc HTtag_/blockquote {param text} {
	upvar #0 HTarr var
	incr var(bqlvl) -1
	set var(indent) [HTcalcindent]
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
}

proc HTtag_address {param text} {
	upvar #0 HTarr var
	incr var(bqlvl)
	set var(indent) [HTcalcindent]
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
}

proc HTtag_/address {param text} {
	upvar #0 HTarr var
	incr var(bqlvl) -1
	set var(indent) [HTcalcindent]
	HTnewpara $var(TextWidth) $var(indent) $var(indent) ""
}

# list element tags

proc HTtag_ol {param text} {
	upvar #0 HTarr var
	incr var(listlvl)
	set var(indent) [HTcalcindent]
	set var(count$var(listlvl)) 0
}

proc HTtag_/ol {param text} {
	upvar #0 HTarr var
	incr var(listlvl) -1
	set var(indent) [HTcalcindent]
}

proc HTtag_ul {param text} {
	upvar #0 HTarr var
	incr var(listlvl)
	set var(indent) [HTcalcindent]
	catch {unset var(count$var(listlvl))}
}

proc HTtag_/ul {param text} {
	upvar #0 HTarr var
	incr var(listlvl) -1
	set var(indent) [HTcalcindent]
}

proc HTtag_menu {param text} {
	upvar #0 HTarr var
	incr var(listlvl)
	set var(indent) [HTcalcindent]
	set var(menu) ->
	set var(compact) 1
}

proc HTtag_/menu {param text} {
	upvar #0 HTarr var
	incr var(listlvl) -1
	set var(indent) [HTcalcindent]
	catch {unset var(menu)}
	catch {unset var(compact)}
}
	
proc HTtag_dl {param text} {
	upvar #0 HTarr var
	incr var(listlvl)
	set var(indent) [HTcalcindent]
}

proc HTtag_/dl {param text} {
	upvar #0 HTarr var
	incr var(listlvl) -1
	set var(indent) [HTcalcindent]
	catch {unset var(dt)}
}

proc HTtag_dt {param text} {
	upvar #0 HTarr var
	set var(dt) 1
	HTnewpara $var(TextWidth) $var(indent) [expr $var(indent)+$var(DTwidth)] ""
}

proc HTtag_dd {param text} {
	upvar #0 HTarr var
	upvar $text data
	set dttext [string trim $var(para)]
	set dtlen  [expr [string length $dttext]+1]
	if { ![info exists var(dt)] || $dtlen >= $var(DTwidth) } {
		set indent [expr $var(indent)+$var(DTwidth)]
		HTnewpara $var(TextWidth) $indent $indent ""
	} else {
		set var(dttext) $dttext
		set var(para) ""
	}
}

proc HTtag_li {param text} {
	upvar #0 HTarr var
	upvar $text data
	set level $var(listlvl)
	set x [string index $var(Symbols) [expr $level-1]]
	catch {set x [incr var(count$level)]}
	catch {set x $var(menu)}
	HTnewpara $var(TextWidth) $var(indent) \
		[expr $var(indent)+[string length $x]+1] "$x $data"
	set data {}
}

# hypertext links.

proc HTtag_a {param text} {
	upvar $text data

	set data [string trimleft $data]
}

proc HTtag_/a {param text} {
	upvar #0 HTarr var
	upvar $text data

	set var(para) [string trimright $var(para)]
}

proc HTrender {tag not param text} {
	global HMinsert_map HTarr HTtext_map
	set tag [string tolower $tag]
	set text [HMmap_esc $text]

	# do any special tag processing
	catch {HTtag_$not$tag $param text} msg

	if {[info exists HMinsert_map($not$tag)] && !$HTarr(nonewline) } {
		set nl $HMinsert_map($not$tag)
		if {[string length $nl] > 1} {
			append HTarr(out) [string range $nl 1 end]
		}
	}
	if { "X$not$tag" != "X/pre" } {
		set HTarr(nonewline) 0
	}
	    
	if {[info exists HTtext_map($tag)]} {
	    if { "X$not" == "X/" } {
		set idx end
	    } else {
		set idx 0
	    }
	    if $HTarr(preflag) {
		append HTarr(out) [lindex $HTtext_map($tag) $idx]
	    } else {
		append HTarr(para) [lindex $HTtext_map($tag) $idx]
	    }
	}
	    
	if [string length $text] {
	    if $HTarr(preflag) {
		append HTarr(out) $text
	    } else {
	        append HTarr(para) "$text"
	    }
	}
}

proc HTnewpara {llen ind1 ind2 text} {
	global HTarr

	set linecount 0
	#puts stderr "PARA:$HTarr(para):$text:"
	set HTarr(para) [string trim [HMzap_white $HTarr(para)]]
	#puts stderr "LLEN =$HTarr(pllen)="
	if {[string length $HTarr(para)] > 0} {
		foreach line [HTparafmt $HTarr(pllen) \
		    $HTarr(pind1) $HTarr(pind2) $HTarr(para)] {
			#puts stderr "FORMATTING :$line:" 
			append HTarr(out) "$line\n"
			incr linecount
		}
	}
	set HTarr(pllen) $llen
	set HTarr(pind1) $ind1
	set HTarr(pind2) $ind2
	set HTarr(para)  $text
	return $linecount
}

proc HTparafmt { llen indent1 indent2 string } {
	#puts stderr "PARAFMT =$llen=$indent1=$indent2=$string="
	upvar #0 HTarr var
	if { $llen <= $indent1 || $llen <= $indent2
		|| $llen < 1 || $indent1 < 0 || $indent2 < 0 } {
			error "Invalid values given - linelength must be at least 1;\nindent1 and indent2 must be at least 0 and must be less than linelength"
	}
	#puts stderr "PARAFMT - 2"
	if { [info exists var(dttext)] } {
		set tmp [string trim [HMzap_white $var(dttext)]]
		append tmp [replicate " " \
			[expr $var(DTwidth)-[string length $tmp]]]
		append tmp $string
		set string $tmp
		catch {unset var(dttext)}
	}
	#puts stderr "PARAFMT - 3"
	set indent $indent1
	set retlist ""
	while { [string length $string] > $llen-$indent } {
		#puts stderr "-- retlist = :$retlist:"
		set lind [expr $llen-$indent]
		set blanks [replicate " " $indent]
		set indent $indent2
		set tmp [string range $string 0 $lind]
		set pos [string last " " $tmp]
		if { $pos == -1 } {
			lappend retlist $blanks[string range $string \
						0 [expr $lind-1]]
			set string [string range $string $lind end]
			continue
		}
		if { $pos == 0 } {
			lappend retlist $blanks[string range $string \
						1 [expr $lind-1]] 
			set string [string range $string $lind end]
			continue
		}
		if { $pos == $lind-1 } {
			lappend retlist $blanks[string range $string 0 \
						[expr $lind-2]]
			set string [string range $string $lind end]
			continue
		}
		lappend retlist $blanks[string range $tmp 0 [expr $pos-1]]
		set string [string range $string [expr $pos+1] end]
	}
	if { [llength $retlist] == 1 } {
		set indent $indent2
	}
	return [lappend retlist [replicate " " $indent][string trim $string]]
}

proc HTcalcindent {} {
	global HTarr
	return [expr $HTarr(baseindent)+2*$HTarr(listlvl)+3*$HTarr(bqlvl)]
}

# if we're not running under TclX, use this substitute proc
if { [info commands replicate] != "replicate" } {
	proc replicate {str count} {
		set retval ""
		for {set i 1} {$i <= $count} {incr i} {
			append retval $str
		}
		return $retval
	}
}

proc HTconvert html {
	global HTarr
	regsub -all "\\\\\n" $html "\\\&#92;\n" html
	HMparse_html $html HTrender
	return $HTarr(out)
}

